"""Test cases for features defined on glue.py module.

This file handles the definition of all test cases for testing GlueJobManager
class and its features. The idea is to isolate a test script for testing
sparksnake features delivered for users who want to develop Spark applications
using AWS Glue environment through a SparkETLManager class object with
mode="glue".

___
"""

# Importing libraries
import pytest
import logging

from sparksnake.manager import SparkETLManager

from pyspark.context import SparkContext
from pyspark.sql import SparkSession

from awsglue.context import GlueContext
from awsglue.job import Job


@pytest.mark.spark_manager_glue
@pytest.mark.constructor
def test_error_on_using_glue_mode_and_not_providing_essential_attributes():
    """
    G: Given that an user wants to initialize a Glue job on AWS
    W: When the class SparkETLManager is initialized with mode="glue" but
       without argv_list or data_dict attributes on object construction
    T: Then a TypeError exception must be raised
    """

    # Trying to initialize SparkETLManager with glue mode and without argv_list
    with pytest.raises(TypeError):
        _ = SparkETLManager(mode="glue")


@pytest.mark.spark_manager_glue
@pytest.mark.constructor
def test_args_attribute_for_class_on_glue_mode_must_be_dict(
    spark_manager_glue
):
    """
    G: Given that an user wants to initialize a Glue job on AWS
    W: When an object of class SparkETLManager is initialized with
       mode="glue"
    T: Then the self.args class attribute must be a Python dictionary
    """

    assert type(spark_manager_glue.args) is dict


@pytest.mark.spark_manager_glue
@pytest.mark.constructor
def test_args_attribute_contains_user_defined_arguments(
    spark_manager_glue
):
    """
    G: Given that a user wants to initialize a Glue job on AWS
    W: When the class SparkETLManager is initialized with mode="glue"
    T: Then the arguments passed on argv_list parameter must be contained on
       self.args class attribute
    """

    # Getting argument job names
    job_arg_names = list(spark_manager_glue.args.keys())

    assert all(a in job_arg_names for a in spark_manager_glue.argv_list)


@pytest.mark.spark_manager_glue
@pytest.mark.job_initial_log_message
def test_content_of_initial_log_message_match_the_expected(
    spark_manager_glue,
    caplog
):
    """
    G: Given that a user wants to initialize a Glue job on AWS
    W: When the job_initial_message() is called from SparkETLManager with
       glue mode
    T: Then the log message captured must match a expected value
    """

    # Defining the expected log message based on data_dict_dictionary used
    expected_log_msg = "Initializing the execution of a-fake-arg-value job. "\
        "Data sources used in this ETL process:\n\n"\
        "Table some-fake-database.orders-fake-table without push down "\
        "predicate\n"\
        "Table some-fake-database.customers-fake-table with the following "\
        "push down predicate info: anomesdia=20221201\n"

    # Calling the method for logging the initial message
    with caplog.at_level(logging.INFO):
        spark_manager_glue.job_initial_log_message()

    assert caplog.records[-1].message == expected_log_msg


@pytest.mark.spark_manager_glue
@pytest.mark.print_args
def test_print_args_method_really_prints_job_arguments_defined_by_user(
    spark_manager_glue,
    caplog
):
    """
    G: Given that a user wants to initialize a Glue job on AWS
    W: When the print_args() is called from SparkETLManager with glue mode
    T: Then the log message captured must have the arguments defined by user
       on self.argv_list class attribute
    """

    # Creating a list for the expected message based on user defined job args
    expected_argv_msg = [f'--{arg}="a-fake-arg-value"'
                         for arg in spark_manager_glue.argv_list]

    # Calling the method
    with caplog.at_level(logging.INFO):
        spark_manager_glue.print_args()

    assert all(a in caplog.text.split("\n") for a in expected_argv_msg)


@pytest.mark.spark_manager_glue
@pytest.mark.get_context_and_session
def test_get_context_and_session_method_generates_context_and_session(
    spark_manager_glue
):
    """
    G: Given that a user wants to get the context and session elements
    W: When method get_context_and_session() is called from SparkETLManager
       with glue mode
    T: Then the object class must have new attributes identified as
       self.sc, self.glueContext and self.spark
    """

    # Calling the method for getting context and session elements
    spark_manager_glue.get_context_and_session()

    # Getting class attributes and defining an assertion list
    class_attribs = spark_manager_glue.__dict__
    attribs_names = ["sc", "glueContext", "spark"]

    # Asserting class has now new attributes
    assert all(a in list(class_attribs.keys()) for a in attribs_names)


@pytest.mark.spark_manager_glue
@pytest.mark.get_context_and_session
def test_correct_type_of_context_and_session_elements_gotten_in_class(
    spark_manager_glue
):
    """
    G: Given that users want to get the context and session elements
    W: When method get_context_and_session() is called from SparkETLManager
       with glue mode
    T: Then the self.sc must be of type SparkContext, the self.glueContext
       attrib must be of type GlueContext and the self.spark attribute must be
       of type SparkSession
    """

    # Calling the method for getting context and session elements
    spark_manager_glue.get_context_and_session()

    # Getting class attributes dictionary
    class_attribs = spark_manager_glue.__dict__

    # Asserting type of each element
    assert type(class_attribs["sc"]) == SparkContext
    assert type(class_attribs["glueContext"]) == GlueContext
    assert type(class_attribs["spark"]) == SparkSession


@pytest.mark.spark_manager_glue
@pytest.mark.init_job
def test_init_job_method_generates_context_and_session_elements(
    spark_manager_glue
):
    """
    G: Given that users want to initialize a Glue job on AWS
    W: When method init_job() is called from SparkETLManager with glue mode
    T: Then context and session elements from Glue and Spark must exist and
       its types must match
    """

    # Calling method for job initialization
    spark_manager_glue.init_job()

    # Getting class attribs and setting up a assertion list
    class_attribs = spark_manager_glue.__dict__
    attribs_names = ["sc", "glueContext", "spark"]

    # Asserting if the new attributes really exist
    assert all(a in list(class_attribs.keys()) for a in attribs_names)

    # Asserting if types match
    assert type(class_attribs["sc"]) == SparkContext
    assert type(class_attribs["glueContext"]) == GlueContext
    assert type(class_attribs["spark"]) == SparkSession


@pytest.mark.spark_manager_glue
@pytest.mark.init_job
def test_init_job_methods_generates_a_job_type_attribute(spark_manager_glue):
    """
    G: Given that users want to initialize a Glue job on AWS
    W: When method init_job() is called from SparkETLManager with glue mode
    T: Then a new class attribute self.job of type awsglue.job.Job must exists
    """

    # Calling method for job initialization
    spark_manager_glue.init_job()

    # Getting class attributes
    class_attribs = spark_manager_glue.__dict__

    assert "job" in class_attribs.keys()
    assert type(class_attribs["job"]) is Job



"""Confest file for managing pytest fixtures and other components.

This file will handle essential components and elements to be used on test
scripts along the project, like features and other things.

___
"""

# Importing libraries
import sys
import pytest

from sparksnake.manager import SparkETLManager
from sparksnake.tester.dataframes import generate_fake_dataframe,\
    generate_dataframe_schema, generate_fake_data_from_schema

from tests.helpers.user_inputs import FAKE_ARGV_LIST, FAKE_DATA_DICT,\
    FAKE_SCHEMA_INFO

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType


# Getting the active SparkSession object (or creating one)
spark = SparkSession.builder.getOrCreate()


# Returning the SparkSession object as a fixture
@pytest.fixture()
def spark_session(spark=spark) -> SparkSession:
    return spark


# A SparkETLManager class object with mode="local"
@pytest.fixture()
def spark_manager_local() -> SparkETLManager:
    return SparkETLManager(mode="local")


# A GlueJobManager class object
@pytest.fixture()
def spark_manager_glue() -> SparkETLManager:
    # Adding system args
    for fake_arg in FAKE_ARGV_LIST:
        sys.argv.append(f"--{fake_arg}=a-fake-arg-value")

    # Initializing a class object
    return SparkETLManager(
        mode="glue",
        argv_list=FAKE_ARGV_LIST,
        data_dict=FAKE_DATA_DICT
    )


# A StructType schema object based on a predefined list of attributes
@pytest.fixture()
def fake_schema() -> StructType:
    return generate_dataframe_schema(schema_info=FAKE_SCHEMA_INFO)


# A list object with fake data generated by generate_fake_data_from_schema()
@pytest.fixture()
def fake_data(fake_schema) -> list:
    return generate_fake_data_from_schema(schema=fake_schema, n_rows=5)


# A fake Spark DataFrame object
@pytest.fixture()
def df_fake(spark_session) -> DataFrame:
    return generate_fake_dataframe(
        spark_session=spark_session,
        schema_info=FAKE_SCHEMA_INFO
    )



"""Centralizing all user inputs for helping on fixtures and test cases.

This file aims to put together all variables used on fixture definitions and
test cases that requires user inputs as a way to configure or validate
something.

The idea behind this file is to have everything related to user inputs on
test cases in a single place. This makes easier to handle, give maintenance,
support and improvements for building new test cases.

___
"""

# A fake argument list for creating Glue jobs
FAKE_ARGV_LIST = ["JOB_NAME", "S3_SOURCE_PATH", "S3_OUTPUT_PATH"]

# A fake data dictionary for setting up data sources
FAKE_DATA_DICT = {
    "orders": {
        "database": "some-fake-database",
        "table_name": "orders-fake-table",
        "transformation_ctx": "dyf_orders"
    },
    "customers": {
        "database": "some-fake-database",
        "table_name": "customers-fake-table",
        "transformation_ctx": "dyf_customers",
        "push_down_predicate": "anomesdia=20221201",
        "create_temp_view": True,
        "additional_options": {
            "compressionType": "lzo"
        }
    }
}

# A fake DataFrame schema object to create Spark DataFrames for test purposes
FAKE_SCHEMA_INFO = [
    {
        "Name": "string_field",
        "Type": "string",
        "nullable": True
    },
    {
        "Name": "integer_field",
        "Type": "int",
        "nullable": True
    },
    {
        "Name": "long_field",
        "Type": "long",
        "nullable": True
    },
    {
        "Name": "decimal_field",
        "Type": "decimal",
        "nullable": True
    },
    {
        "Name": "float_field",
        "Type": "float",
        "nullable": True
    },
    {
        "Name": "double_field",
        "Type": "double",
        "nullable": True
    },
    {
        "Name": "boolean_field",
        "Type": "boolean",
        "nullable": True
    },
    {
        "Name": "date_field",
        "Type": "date",
        "nullable": True
    },
    {
        "Name": "timestamp_field",
        "Type": "timestamp",
        "nullable": False
    }
]



"""Handling operations that help users to improve their test cases.

This module puts together some useful functions created in order to provid
an easy way to fake Spark DataFrames objects. Its features can be imported
and applied on every scenario that demands the creation of fake data rows,
fake schema or even fake Spark DataFrame objects (for example, a conftest
file that defined fixtures for unit test cases).

___
"""

# Importing libraries
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType, StructField, StringType,\
    IntegerType, LongType, DecimalType, FloatType, DoubleType, BooleanType,\
    DateType, TimestampType

from faker import Faker
from decimal import Decimal
from random import random, randrange


# Creating a faker object
faker = Faker()
Faker.seed(42)

# Getting the active SparkSession object (or creating one)
spark = SparkSession.builder.getOrCreate()


# Parsing a string for a dtype into a valid Spark dtype
def parse_string_to_spark_dtype(dtype: str):
    """Transform a string dtype reference into a valid Spark dtype.

    This function checks for the data type reference for a field given by users
    while filling the JSON schema file in order to return a valid Spark dtype
    based on the string reference.

    Examples:
        ```python
        # Returning the Spark reference for a "string" data type
        spark_dtype = parse_string_to_spark_dtype(dtype="string")
        # spark_dtype now holds the StringType Spark dtype object
        ```

    Args:
        dtype (str): A string reference for any parseable Spark dtype

    Returns:
        A callable Spark dtype object based on the string reference provided
    """

    # Removing noise on string before validating
    dtype_prep = dtype.lower().strip()

    # Parsing string reference for dtype to spark data type
    if dtype_prep == "string":
        return StringType
    elif dtype_prep in ("int", "integer"):
        return IntegerType
    elif dtype_prep in ("bigint", "long"):
        return LongType
    elif dtype_prep == "decimal":
        return DecimalType
    elif dtype_prep == "float":
        return FloatType
    elif dtype_prep == "double":
        return DoubleType
    elif dtype_prep == "boolean":
        return BooleanType
    elif dtype_prep == "date":
        return DateType
    elif dtype == "timestamp":
        return TimestampType
    else:
        raise TypeError(f"Data type {dtype} is not valid or currently "
                        "parseable into a native Spark dtype")


# Creating a valid Spark DataFrame schema from a list with fields information
def generate_dataframe_schema(
    schema_info: list,
    attribute_name_key: str = "Name",
    dtype_key: str = "Type",
    nullable_key: str = "nullable"
) -> StructType:
    """Generates a StructType Spark schema based on a list of fields info.

    This function receives a preconfigured Python list extracted from a JSON
    schema definition file provided by user in order to return a valid Spark
    schema composed by a StructType structure with multiple StructField objects
    containing informations about name, data type and nullable info about
    attributes.

    Examples:
        ```python
        # Showing an example of a input schema list
        schema_info = [
            {
                "Name": "idx",
                "Type": "int",
                "nullable": true
            },
            {
                "Name": "order_id",
                "Type": "string",
                "nullable": true
            }
        ]

        # Returning a valid Spark schema object based on a dictionary
        schema = generate_dataframe_schema(schema_info)
        ```

    Args:
        schema_info (list):
            A list with information about fields of a DataFrame

        attribute_name_key (str):
            A string identification of the attribute name defined on every
            attribute dictionary

        dtype_key (str):
            A string identification of the attribute type defined on every
            attribute dictionary

        nullable_key (bool):
            A boolean flag that tells if the given attribute defined in
            the dictionary can hold null values

    Returns:
        A StructType object structured in such a way that makes it possible to\
        create a Spark DataFrame with a predefined schema.
    """

    # Extracing the schema based on the preconfigured dict info
    schema = StructType([
        StructField(
            field_info[attribute_name_key],
            parse_string_to_spark_dtype(field_info[dtype_key])(),
            nullable=field_info[nullable_key]
            if nullable_key in field_info.keys() else True
        ) for field_info in schema_info
    ])

    return schema


# Generating fake data based on native Spark data types and the Faker library
def generate_fake_data_from_schema(
    schema: StructType,
    n_rows: int = 5
) -> tuple:
    """Generates fake data based on a Struct Type Spark schema object.

    This function receives a predefined DataFrame schema in order to return
    a list of tuples with fake data generated based on attribute types and
    the Faker library. The way the fake data is structured makes it easy to
    create Spark DataFrames to be used for test purposes.

    Examples:
        ```python
        # Defining a list with attributes info to be used on schema creation
        schema_info = [
            {
                "Name": "idx",
                "Type": "int",
                "nullable": true
            },
            {
                "Name": "order_id",
                "Type": "string",
                "nullable": true
            }
        ]

        # Returning a valid Spark schema object based on a dictionary
        schema = generate_dataframe_schema(schema_info)

        # Generating fake data based on a Spark DataFrame schema
        fake_data = generate_fake_data_from_schema(schema=schema, n_rows=10)
        ```

    Args:
        schema (StructType): a Spark DataFrame schema
        n_rows (int): the number of fake rows to be generated

    Returns:
        A list of tuples where each tuple representes a row with fake data\
        generated using the Faker library according to each data type of\
        the given Spark DataFrame schema. For example, for a string attribute\
        the fake data will be generated using the `faker.word()` method. For a\
        date attribute, the fake data will be generated using the\
        `faker.date_this_year()`. And so it goes on for all other dtypes.
    """

    # Creating fake data based on each schema attribute
    fake_data_list = []
    for _ in range(n_rows):
        # Iterting over columns and faking data
        fake_row = []
        for field in schema:
            dtype = field.dataType.typeName()
            if dtype == "string":
                fake_row.append(faker.word())
            elif dtype in ("int", "integer"):
                fake_row.append(randrange(-10000, 10000))
            elif dtype in ("bigint", "long"):
                fake_row.append(randrange(-10000, 10000))
            elif dtype == "decimal":
                fake_row.append(Decimal(randrange(1, 100000)))
            elif dtype in ("float", "double"):
                fake_row.append(float(random() * randrange(1, 100000)))
            elif dtype == "boolean":
                fake_row.append(faker.boolean())
            elif dtype == "date":
                fake_row.append(faker.date_this_year())
            elif dtype == "timestamp":
                fake_row.append(faker.date_time_this_year())

        # Appending the row to the data list
        fake_data_list.append(fake_row)

    # Generating a list of tuples
    return [tuple(row) for row in fake_data_list]


# Generating Spark DataFrame objects with fake data
def generate_fake_dataframe(
    spark_session: SparkSession,
    schema_info: list,
    attribute_name_key: str = "Name",
    dtype_key: str = "Type",
    nullable_key: str = "nullable",
    n_rows: int = 5
) -> DataFrame:
    """Creates a Spark DataFrame with fake data using Faker.

    This function receives a list of dictionaries, each one populated with
    information about the desired attributes defined in order to create a
    Spark DataFrame with fake data. So, this list of dictionaries (schema_info
    function argument) is used to create a StructType Spark DataFrame schema
    object and this objects is used to generate fake data using Faker and based
    on the type of the attributes defined on the schema. Finally, with the
    schema object and the fake data, this function returns a Spark DataFrame
    that can be used for any purposes.

    This function calls the generate_dataframe_schema() and
    generate_fake_data_from_schema() in order to execute all the the steps
    explained above.

    Examples:
        ```python
        # Defining a list with attributes info to be used on schema creation
        schema_info = [
            {
                "Name": "idx",
                "Type": "int",
                "nullable": true
            },
            {
                "Name": "order_id",
                "Type": "string",
                "nullable": true
            }
        ]

        # Generating a Spark DataFrame object with fake data
        fake_df = generate_fake_dataframe(schema_info)
        ```

    Args:
        spark_session (SparkSession):
            A SparkSession object that is used to call createDataFrame method

        schema_info (list):
            A list with information about fields of a DataFrame. Check the
            generate_dataframe_schema() for more details.

        attribute_name_key (str):
            A string identification of the attribute name defined on every
            attribute dictionary. Check the generate_dataframe_schema() for
            more details.

        dtype_key (str):
            A string identification of the attribute type defined on every
            attribute dictionary. Check the generate_dataframe_schema() for
            more details.

        nullable_key (bool):
            A boolean flag that tells if the given attribute defined in
            the dictionary can hold null values. Check the
            generate_dataframe_schema() for more details.

        n_rows (int):
            The number of fake rows to be generated. Check the
            generate_fake_data_from_schema() for more details.

    Returns:
        A new Spark DataFrame with fake data generated by Faker providers and
        Python built-in libraries.
    """

    # Returning a valid Spark schema object based on a dictionary
    schema = generate_dataframe_schema(
        schema_info=schema_info,
        attribute_name_key=attribute_name_key,
        dtype_key=dtype_key,
        nullable_key=nullable_key
    )

    # Generating fake data based on a Spark DataFrame schema
    fake_data = generate_fake_data_from_schema(schema=schema, n_rows=n_rows)

    # Returning a fake Spark DataFrame
    return spark_session.createDataFrame(data=fake_data, schema=schema)


# Comparing Spark schemas based on custom conditions
def compare_dataframe_schemas(
    df1: DataFrame,
    df2: DataFrame,
    compare_nullable_info: bool = False
) -> bool:
    """Compares the schema from two Spark DataFrames with custom options.

    This function helps users to compare two Spark DataFrames schemas based on
    custom conditions provided in order to help the comparison.

    The schema of a Spark DataFrame is made of three main elements:
    column name, column type and a boolean information telling if the field
    accepts null values. In some cases, this third element can cause errors
    when comparing two DataFrame schemas. Imagine that a Spark DataFrame is
    created from a transformation function and there is no way to configure
    if a field accepts a null value without (think of an aggregation step that
    can create null values for some rows... or not). So, when comparing schemas
    from two DataFrames, maybe we are interested only on column names and data
    types, and not if an attribute is nullable or not.

    This function uses a flag in order to provide two ways to compare Spark
    DataFrame schemas:
        1. Removing the "nullable" info from schema
        2. Comparing native DataFrame.schema attributes without any changes

    Examples:
        ```python
        compare_dataframe_schemas(df1, df2, compare_nullable_info=False)
        # Result is True or False
        ```

    Args:
        df1 (pyspark.sql.DataFrame): the first Spark DataFrame to be compared
        df2 (pyspark.sql.DataFrame): the second Spark DataFrame to be compared
        compare_nullable_info (bool):
            a boolean flag that leads to compare the schemas including the
            nullable information or not.

    Return:
        The function returns True if both DataFrame schemas are equal or\
        False if it isn't.
    """

    # Extracting infos to be compared based on user conditions
    if not compare_nullable_info:
        df1_schema = [[col.name, col.dataType] for col in df1.schema]
        df2_schema = [[col.name, col.dataType] for col in df2.schema]
    else:
        df1_schema = df1.schema
        df2_schema = df2.schema

    # Checking if schemas are equal
    return df1_schema == df2_schema



"""Managing common operations found in ETL jobs developed using AWS Glue.

This module aims to put together everything it's needed to improve and enhance
the development journey of Glue jobs on AWS.

___
"""

# Importing libraries
import sys
from time import sleep
from sparksnake.utils.log import log_config

from awsglue.utils import getResolvedOptions
from awsglue.context import GlueContext
from awsglue.job import Job
from awsglue.dynamicframe import DynamicFrame

from pyspark.context import SparkContext
from pyspark.sql import DataFrame


# Setting up a logger object
logger = log_config(logger_name=__file__)


class GlueJobManager():
    """Management of elements and operations commonly used in Glue jobs.

    This class is responsible for managing and providing all necessary inputs
    for launching and developing Glue jobs in AWS. All attributes and methods
    declared in this class are *inherited* by `SparkETLManager`class in the
    `manager` module.

    To see a end to end usage example, check the docs for the `SparkETLManager`
    class.

    Args:
        argv_list (list):
            List with all user defined arguments for the job

        data_dict (dict):
            Dictionary with all source data references to be read from the
            catalog and used in the ETL job.

    This class can get some other attributes along the execution of its
    methods. Those are:

    Attributes:
        args (dict):
            Dictionary with all arguments of the job. This is a joint
            between user defined arguments and system arguments (`sys.argv`)

        sc (SparkContext):
            A Spark context object used for creating a Glue context object

        glueContext (GlueContext):
            A Glue context object used for creating a Spark session object

        spark (SparkSession):
            A Spark session object used as a central point for job operations

    Tip: About setting up the data_dict class attribute
        The data_dict dictionary passed as a required attribute for the
        `GlueJobManager` class must be defined following some rules.

        Your main purpose is to provide a single point for controlling all
        data sources used in the job. So, your composition is something really
        important for ensuring that reading process can be executed in an
        expected way.

        With that in mind, it's crucial to say that the data_dict dictionary
        can be defined with all parameters found in
        `glueContext.create_dynamic_frame.from_catalog` Glue method. In other
        words, all keys of data_dict dictionary can be assume any valid
        parameter of the Glue method mentioned above. Of course the values
        of those keys defined in data_dict depends on the job rules.

        :star: It means that if the user wants to read a data source from the
        catalog using the push down predicate Glue feature, it's only necessary
        to define a key called "push_down_predicate" (the same way as found in
        `glueContext.create_dynamic_frame.from_catalog` method) with the value
        to be filtered. This is the same rule for all other parameters, such as
        additional_options, catalog_id, transformation_ctx, and others. If the
        user wants to read some data source with a special parameter, just take
        a look at the mentioned Glue method above and input a key on the
        data_dict to set up the data source reading process.

        By the other hand, it's also important to mention that the user is not
        forced to define the data_dict dictionary with all parameters found in
        `glueContext.create_dynamic_frame.from_catalog` method. If the
        data_dict attribute doesn't have a key for a specific parameter on the
        Glue method mentioned abouve, so its default value will be considered.
        For example, that if data_dict doesn't have a push_down_predicate key
        for a data source, the value "None" will be considered as it's the
        default value for this key on the
        `glueContext.create_dynamic_frame.from_catalog` method.
    """

    def __init__(self, argv_list: list, data_dict: dict) -> None:
        self.argv_list = argv_list
        self.data_dict = data_dict

        # Getting job arguments
        self.args = getResolvedOptions(sys.argv, self.argv_list)

    def job_initial_log_message(self) -> None:
        """Preparation of a detailed log message for job initialization.

        This method is responsible for composing an initial log message to be
        shown in CloudWatch after the user initializes a Glue Job. The message
        aim to clarify some job details, such as all the data sources mapped
        and its push down predicate values (if used). This can be a good
        practice for making Glue jobs cleaner and more organized.
        """

        # Defining a initial string for composing the message
        welcome_msg = f"Initializing the execution of {self.args['JOB_NAME']}"\
                      " job. Data sources used in this ETL process:\n\n"
        initial_msg = ""

        # Iterating over the data_dict dicionary for extracting some info
        for _, params in self.data_dict.items():
            # Getting table name and start building the message
            tbl_ref = f"{params['database']}.{params['table_name']}"
            table_msg = f"Table {tbl_ref} "

            # Looking for a push_down_predicate information in the dictionary
            if "push_down_predicate" in params:
                table_msg += "with the following push down predicate info: "\
                             f"{params['push_down_predicate']}\n"
            else:
                table_msg += "without push down predicate\n"

            # Concat final message
            initial_msg += table_msg

        # Logging the message to CloudWatch
        logger.info(welcome_msg + initial_msg)

    def print_args(self) -> None:
        """Getting and logging job arguments in CloudWatch.

        This method is responsible to show users all the arguments used in the
        job. It achieves its goal by iterating over all args in self.args class
        attribut for composing a detailed log message.
        """

        # Formatting and composing a log message with job arguments
        args_formatted = "".join([f'--{k}="{v}"\n'
                                  for k, v in self.args.items()])
        logger.info(f"Job arguments:\n\n{args_formatted}")
        sleep(0.01)

    def get_context_and_session(self) -> None:
        """Getting context and session elements for Glue job application.

        This method is a central point for creating the following elements:
        SparkContext, GlueContext and SparkSession.
        """
        logger.info("Creating SparkContext, GlueContext and SparkSession")
        self.sc = SparkContext.getOrCreate()
        self.glueContext = GlueContext(self.sc)
        self.spark = self.glueContext.spark_session

    def init_job(self):
        """Initializing a Glue job.

        This method consolidates all the necessary steps required for a Glue
        job initialization process. In its definition, it calls the following
        methods:

        * `self.job_initial_log_message()`
        * `self.print_args()`
        * `self.get_context_and_session()`

        After that, the class has all the attributes required for calling the
        Job class used by Glue to initialize a job object.
        """

        # Getting job arguments and logging useful infos for users
        self.job_initial_log_message()
        self.print_args()

        # Getting context and session elements
        self.get_context_and_session()

        # Initializing a Glue job based on a GlueContext element
        self.job = Job(self.glueContext)
        self.job.init(self.args['JOB_NAME'], self.args)

    def generate_dynamicframes_dict(self) -> dict:
        """Getting a dictionary of DynamicFrames objects read from the catalog.

        This method uses the data_dict class attribute for reading and getting
        DynamicFrame objects mapped as data sources on the mentioned
        dictionary. The main advantage of using this method is explained by
        having the possibility to read multiple data sources with a single
        method call. The result of this method is a dictionary containing all
        DynamicFrames objects as values of keys mapped on data_dict class
        attribute. The user will be able to access those DynamicFrames in
        a easy way through indexing.

        Examples:
            ```python
            # Getting a dictionary of Glue DynamicFrames
            dyfs_dict = spark_manager.generate_dynamicframes_dict()

            # Inexing and getting individual DynamicFrames
            dyf_orders = dyfs_dict["orders"]
            dyf_customers = dyfs_dict["customers"]
            ```

        Returns:
            Python dictionary with keys representing the identification of the\
            data source put into the data_dict class attributes and values\
            representing DynamicFrame objects read from catalog.

        Tip: Details about the returned dictionary from the method
            In order to provide a clear view of the return of this method,
            consider the following definition example of `self.data_dict` as a
            class attribute:

            ```python
            {
                "orders": {
                    "database": "ra8",
                    "table_name": "orders",
                    "transformation_ctx": "dyf_orders"
                },
                "customers": {
                    "database": "ra8",
                    "table_name": "customers",
                    "transformation_ctx": "dyf_customers",
                    "push_down_predicate": "anomesdia=20221201",
                    "create_temp_view": True,
                    "additional_options": {
                        "compressionType": "lzo"
                    }
                }
            }
            ```

            As stated before, all
            `glueContext.create_dynamic_frame.from_catalog()` can be used as
            dictionary keys on `self.data_dict` definition. Moreover, some
            additional keys can be defined by the user for some special
            purposes, such as:

            - "create_temp_view": bool -> Sets the creation of a Spark
                temporary table (view) after reading the data source as a
                DynamicFrame

            The return of the method `generate_dynamicframes_dict()` will be
            presented as the following format:

            ```python
            {
                "orders": <DynamicFrame>
                "customers": <DynamicFrame>
            }
            ```

            where the tags <DynamicFrame> means objects of DynamicFrame type
            read for each data source.
        """

        logger.info("Iterating over data_dict dictionary for reading data "
                    "sources as Glue DynamicFrame objects")
        try:
            dynamic_frames = []
            for t in self.data_dict.keys():
                # Getting some required args: database, table_name, ctx
                database = self.data_dict[t]["database"]
                table_name = self.data_dict[t]["table_name"]
                transformation_ctx = self.data_dict[t]["transformation_ctx"]

                # Getting non required args: push_down_predicate
                push_down_predicate = self.data_dict[t]["push_down_predicate"]\
                    if "push_down_predicate" in self.data_dict[t].keys()\
                    else ""

                # Getting non required args: additional_options
                additional_options = self.data_dict[t]["additional_options"] \
                    if "additional_options" in self.data_dict[t].keys()\
                    else {}

                # Getting non required args: catalog_id
                catalog_id = self.data_dict[t]["catalog_id"] \
                    if "catalog_id" in self.data_dict[t].keys()\
                    else None

                # Reading the DynamicFrame
                dyf = self.glueContext.create_dynamic_frame.from_catalog(
                    database=database,
                    table_name=table_name,
                    transformation_ctx=transformation_ctx,
                    push_down_predicate=push_down_predicate,
                    additional_options=additional_options,
                    catalog_id=catalog_id
                )

                # Appending the DynamicFrame obnject to DynamicFrames list
                dynamic_frames.append(dyf)

        except Exception as e:
            logger.error("Error on generating a list of DynamicFrames list. "
                         f"Exception: {e}")
            raise e

        logger.info("Mapping DynamicFrames objects to a dictionary key")
        sleep(0.01)

        # Creating a DynamicFrames dictionary
        dynamic_dict = {k: dyf for k, dyf
                        in zip(self.data_dict.keys(), dynamic_frames)}
        logger.info("Success on reading data. There are "
                    f"{len(dynamic_dict.values())} DynamicFrames available.")

        # Returning the dictionary
        sleep(0.01)
        return dynamic_dict

    def generate_dataframes_dict(self) -> dict:
        """Getting a dictionary of DataFrames objects read from the catalog.

        This method uses the data_dict class attribute for reading and getting
        DataFrame objects mapped as data sources on the mentioned
        dictionary. The main advantage of using this method is explained by
        having the possibility to read multiple data sources with a single
        method call. The result of this method is a dictionary containing all
        DataFrames objects as values of keys mapped on data_dict class
        attribute. The user will be able to access those DataFrames in
        a easy way through indexing.

        In its behalf, this method calls `generate_dynamicframes_dict()` method
        for getting a dictionary of DynamicFrames after applying the method
        `toDF()` for transforming all objects into Spark DataFrames.

        Examples:
            ```python
            # Getting a dictionary of Spark DataFrames
            dyfs_dict = spark_manager.generate_dataframes_dict()

            # Inexing and getting individual DataFrames
            dyf_orders = dyfs_dict["orders"]
            dyf_customers = dyfs_dict["customers"]
            ```

        Returns:
            Python dictionary with keys representing the identification of the\
            data source put into the data_dict class attributes and values\
            representing DataFrame objects read from catalog.

        Tip: Details about the returned dictionary from the method
            In order to provide a clear view of the return of this method,
            consider the following definition example of `self.data_dict` as a
            class attribute:

            ```python
            {
                "orders": {
                    "database": "ra8",
                    "table_name": "orders",
                    "transformation_ctx": "dyf_orders"
                },
                "customers": {
                    "database": "ra8",
                    "table_name": "customers",
                    "transformation_ctx": "dyf_customers",
                    "push_down_predicate": "anomesdia=20221201",
                    "create_temp_view": True,
                    "additional_options": {
                        "compressionType": "lzo"
                    }
                }
            }
            ```

            As stated before, all
            `glueContext.create_dynamic_frame.from_catalog()` can be used as
            dictionary keys on `self.data_dict` definition. Moreover, some
            additional keys can be defined by the user for some special
            purposes, such as:

            - "create_temp_view": bool -> Sets the creation of a Spark
                temporary table (view) after reading the data source as a
                DataFrame

            The return of the method `generate_dataframes_dict()` will be
            presented as the following format:

            ```python
            {
                "orders": <DataFrame>
                "customers": <DataFrame>
            }
            ```

            where the tags <DataFrame> means objects of DataFrame type
            read for each data source.
        """

        # Getting a DynamicFrame objects dictionary
        dyf_dict = self.generate_dynamicframes_dict()

        # Transforming DynamicFrames into DataFrames
        logger.info(f"Transforming {len(dyf_dict.keys())} "
                    "Glue DynamicFrames into Spark DataFrames")
        try:
            df_dict = {k: dyf.toDF() for k, dyf in dyf_dict.items()}
            logger.info("Success on generating all Spark DataFrames")
            sleep(0.01)

        except Exception as e:
            logger.error("Error on transforming Glue DataFrames into Spark "
                         f"DataFrames. Exception: {e}")
            raise e

        # Creating spark temporary tables if applicable
        for table_key, params in self.data_dict.items():
            try:
                # Extracting useful variables from the dict
                df = df_dict[table_key]
                table_name = params["table_name"]

                # Creating temp tables (if the flag for it was set as True)
                if "create_temp_view" in params\
                        and bool(params["create_temp_view"]):
                    df.createOrReplaceTempView(table_name)

                    logger.info(f"Spark temporary table (view) {table_name} "
                                "was successfully created.")

            except Exception as e:
                logger.error("Error on creating Spark temporary table "
                             f"{table_name}. Exception: {e}")
                raise e

        # Returning the DataFrames dict
        sleep(0.01)
        return df_dict

    def drop_partition(self, s3_partition_uri: str,
                       retention_period: int = 0) -> None:
        """Deleting (purging) a physical table partition directly on S3.

        This methods is responsable for excluding physical partitions on S3
        through purge method from a GlueContext. The users can use this feature
        for ensuring that new writing processes in a specific partitions will
        only be done after deleting existing references for the given
        partition.

        Examples:
            ```python
            # Dropping a physycal partition on S3
            partition_uri = "s3://some-bucket/some-table/partition=value/"
            spark_manager.drop_partition(partition_uri)

            # The result is the elimination of everything under the prefix
            # partition=name (including the prefix itself and all table files)
            ```

        Args:
            s3_partition_uri (str): Partition URI on S3
            retention_period (int): Hours to data retention

        Raises:
            Exception: Generic exception raises when a failed attempt of
            executing the `glueContext.purge_s3_path()` method.
        """

        logger.info(f"Searching for the partition {s3_partition_uri} and "
                    "dropping it (if exists)")
        try:
            self.glueContext.purge_s3_path(
                s3_path=s3_partition_uri,
                options={"retentionPeriod": retention_period}
            )
        except Exception as e:
            logger.error(f"Error on purging partition {s3_partition_uri}. "
                         f"Exception: {e}")
            raise e

    def write_and_catalog_data(
            self,
            df: DataFrame or DynamicFrame,
            s3_table_uri: str,
            output_database_name: str,
            output_table_name: str,
            partition_name: str = None,
            connection_type: str = "s3",
            update_behavior: str = "UPDATE_IN_DATABASE",
            compression: str = "snappy",
            enable_update_catalog: bool = True,
            output_data_format: str = "parquet"
    ) -> None:
        """Writing data on S3 anda cataloging on Data Catalog.

        This methods is responsible to put together all the steps needed to
        write a Spark DataFrame or a Glue DynamicFrame into S3 and catalog
        its metadata on Glue Data Catalog. In essence, this methods include
        the following steps:

        1. Check if the data object type (df argument) is a Glue DynamicFrame
        (if it's not, it converts it)
        2. Make a sink with data catalog
        3. Write data in S3 and update the data catalog with the behavior
        chosen by the user

        Examples:
            ```python
            # Writing and cataloging data
            spark_manager.write_data(
                df=df_orders,
                s3_table_uri="s3://some-bucket/some-table",
                output_database_name="db_corp_business_inteligence",
                output_table_name="tbl_orders_prep",
                partition_name="anomesdia"
            )
            ```

        Args:
            df (DataFrame or DynamicFrame):
                A data object that can be a Spark DataFrame
                (`pyspark.sql.DataFrame`) or a Glue DynamicFrame
                (`awsglue.dynamicframe.DynamicFrame`) considered as a target
                element for writing and cataloging processes. This information
                is used on `glueContext.getSink().writeFrame()` method.

            s3_table_uri (str):
                Table URI in the format `s3://bucket-name/table-prefix/`. This
                information is used on parameter "path" from
                `glueContext.getSink()` method.

            output_database_name (str):
                Reference for the database used on the catalog process for the
                table. This information is used on parameter "catalogDatabase"
                from `glueContext.getSink().setCatalogInfo()` method.

            output_table_name (str):
                Reference for the table used on the catalog process. This
                information is used on parameter "catalogTableName"
                from `glueContext.getSink().setCatalogInfo()` method.

            partition_name (str or None):
                Partition coumn name chosen for the table. This information is
                used on parameter "partitionKeys" from `glueContext.getSink()`
                method.

            connection_type (str):
                Connection type used in storage. This information is used on
                parameter "connection_type" from `glueContext.getSink()` method

            update_behavior (str):
                Defines the update behavior for the target table. This
                information is used on parameter "updateBehavior" from
                `glueContext.getSink()` method.

            compression (str):
                Defines the compression method used on data storage process.
                This information is used on parameter "compression" from
                `glueContext.getSink()` method.

            enable_update_catalog (bool):
                Enables the update of data catalog with data storage by this
                method. This information is used on parameter
                "enableUpdateCatalog" from  `glueContext.getSink()` method.

            output_data_format (str):
                Defines the data format for the data to be stored on S3. This
                information is used on parameter "output_data_format" from
                `glueContext.getSink().setFormat()` method.
        """

        # Converting DataFrame into DynamicFrame
        if type(df) == DataFrame:
            logger.info("Converting the Spark DataFrame as Glue DynamicFrame")
            try:
                dyf = DynamicFrame.fromDF(df, self.glueContext, "dyf")
            except Exception as e:
                logger.error("Error on converting the DataFrame object as "
                             f"DynamicFrame. Exception: {e}")
                raise e
        else:
            dyf = df

        # Creating a sink with S3
        logger.info("Creating a sink with all configurations provided")
        try:
            # Criando relação de escrita de dados
            data_sink = self.glueContext.getSink(
                path=s3_table_uri,
                connection_type=connection_type,
                updateBehavior=update_behavior,
                partitionKeys=[partition_name],
                compression=compression,
                enableUpdateCatalog=enable_update_catalog,
                transformation_ctx="data_sink",
            )
        except Exception as e:
            logger.error("Error on creating a sink for the output table. "
                         f"Exception: {e}")
            raise e

        # Setting up data catalog info
        logger.info("Setting up information on the Data Catalog for the table")
        try:
            data_sink.setCatalogInfo(
                catalogDatabase=output_database_name,
                catalogTableName=output_table_name
            )
            data_sink.setFormat(output_data_format,
                                useGlueParquetWriter=True)
            data_sink.writeFrame(dyf)

            logger.info(f"Table {output_database_name}."
                        f"{output_table_name} "
                        "successfully update on Data Catalog. Its data are "
                        f"stored in the following path: {s3_table_uri}")
        except Exception as e:
            logger.error("Error on trying to write data for the given table "
                         f". Exception: {e}")
            raise e



"""Test cases for features defined on tester/dataframes.py module.

This file handles test cases for features on datframes.py module located on
tester parent module, ensuring that everything is working properly.

___
"""

# Importing libraries
import pytest
from decimal import Decimal
from datetime import date, datetime

from sparksnake.tester.dataframes import parse_string_to_spark_dtype,\
    compare_dataframe_schemas, generate_fake_dataframe

from tests.helpers.user_inputs import FAKE_SCHEMA_INFO

from pyspark.sql.types import StringType, IntegerType, LongType, DecimalType,\
    FloatType, DoubleType, BooleanType, DateType, TimestampType, StructType
from pyspark.sql import DataFrame


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_string_reference_correctly_parsed_as_spark_string_dype_object():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument equals to "string"
    T: Then the return must be a StringType object
    """

    assert parse_string_to_spark_dtype("string") == StringType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_int_reference_correctly_parsed_as_spark_integer_dype_object():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument equals to "int"
    T: Then the return must be a IntegerType object
    """

    assert parse_string_to_spark_dtype("int") == IntegerType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_integer_reference_correctly_parsed_as_spark_integer_dype_object():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument equals to "integer"
    T: Then the return must be a IntegerType object
    """

    assert parse_string_to_spark_dtype("integer") == IntegerType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_long_reference_correctly_parsed_as_spark_long_dype_object():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument equals to "long"
    T: Then the return must be a LongType object
    """

    assert parse_string_to_spark_dtype("long") == LongType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_bigint_reference_correctly_parsed_as_spark_long_dype_object():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument equals to "bigint"
    T: Then the return must be a LongType object
    """

    assert parse_string_to_spark_dtype("bigint") == LongType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_decimal_reference_correctly_parsed_as_spark_decimal_dype_object():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument equals to "decimal"
    T: Then the return must be a DecimalType object
    """

    assert parse_string_to_spark_dtype("decimal") == DecimalType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_float_reference_correctly_parsed_as_spark_float_dype_object():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument equals to "float"
    T: Then the return must be a FloatType object
    """

    assert parse_string_to_spark_dtype("float") == FloatType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_double_reference_correctly_parsed_as_spark_double_dype_object():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument equals to "double"
    T: Then the return must be a DoubleType object
    """

    assert parse_string_to_spark_dtype("double") == DoubleType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_boolean_reference_correctly_parsed_as_spark_boolean_dype_object():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument equals to "boolean"
    T: Then the return must be a BooleanType object
    """

    assert parse_string_to_spark_dtype("boolean") == BooleanType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_date_reference_correctly_parsed_as_spark_date_dype_object():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument equals to "date"
    T: Then the return must be a DateType object
    """

    assert parse_string_to_spark_dtype("date") == DateType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_timestamp_reference_correctly_parsed_as_spark_timestamp_dype_object():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument equals to "timestamp"
    T: Then the return must be a TimestampType object
    """

    assert parse_string_to_spark_dtype("timestamp") == TimestampType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.parse_string_to_spark_dtype
def test_error_when_trying_to_parse_an_invalid_string_to_spark_dtype():
    """
    G: Given that users want to get a valid Spark dtype object based on a
       python string reference
    W: When the function parse_string_to_spark_dtype() is called with dtype
       argument assuming a string reference that is not parseable as a valid
       Spark dtype (i.e. "foo")
    T: Then a TypeError exception must be raised
    """

    with pytest.raises(TypeError):
        _ = parse_string_to_spark_dtype(dtype="foo")


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.generate_dataframe_schema
def test_function_generate_dataframe_schema_returns_a_structtype_object(
    fake_schema
):
    """
    G: Given that users want to create a Spark DataFrame schema object using
       a predefined schema list with attributes info
    W: When the funtion generate_dataframe_schema() is called
    T: Then the return object must be a StructType object
    """

    assert type(fake_schema) is StructType


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.generate_dataframe_schema
def test_schema_object_generated_contains_all_predefined_field_names(
    fake_schema
):
    """
    G: Given that users want to create a Spark DataFrame schema object using
       a predefined schema list with attributes info
    W: When the funtion generate_dataframe_schema() is called
    T: Then the schema object returned must have all field names previously
       defined by user on the list used to generate the schema
    """

    # Extracting the field names of the user predefined attributes list
    expected_field_names = [field["Name"] for field in FAKE_SCHEMA_INFO]

    assert fake_schema.fieldNames() == expected_field_names


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.generate_dataframe_schema
def test_schema_object_generated_contains_all_predefined_data_types(
    fake_schema
):
    """
    G: Given that users want to create a Spark DataFrame schema object using
       a predefined schema list with attributes info
    W: When the funtion generate_dataframe_schema() is called
    T: Then the schema object returned must have all data types previously
       defined by user on the list used to generate the schema
    """

    # Extracting the field names of the user predefined attributes list
    expected_dtypes = [
        parse_string_to_spark_dtype(f["Type"]) for f in FAKE_SCHEMA_INFO
    ]

    # Extracting the data types on the returned schema
    schema_dtypes = [type(field.dataType) for field in fake_schema]

    assert schema_dtypes == expected_dtypes


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.generate_fake_data_from_schema
def test_function_generate_fake_data_from_schema_returns_a_list(fake_data):
    """
    G: Given that users want to generate fake data based on a DataFrame schema
    W: When the function generate_fake_data_from_schema() is called
    T: Then the returned object must be a list
    """

    assert type(fake_data) is list


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.generate_fake_data_from_schema
def test_function_generate_fake_data_from_schema_returns_a_list_of_tuples(
    fake_data
):
    """
    G: Given that users want to generate fake data based on a DataFrame schema
    W: When the function generate_fake_data_from_schema() is called
    T: Then all elements for the returned list must be tuples
    """

    # Extracting the type of the inner elements of the fake_data list
    fake_data_inner_elements = [type(row) for row in fake_data]
    fake_data_inner_elements_type = list(set(fake_data_inner_elements))[0]

    assert fake_data_inner_elements_type is tuple


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.generate_fake_data_from_schema
def test_function_generate_fake_data_from_schema_returns_correct_num_of_rows(
    fake_data
):
    """
    G: Given that users want to generate fake data based on a DataFrame schema
    W: When the function generate_fake_data_from_schema() is called with
       n_rows=5 to generate exactly 5 rows
    T: Then the returned object must be a list with 5 elements
    """

    assert len(fake_data) == 5


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.generate_fake_data_from_schema
def test_function_generate_fake_data_from_schema_returns_expected_data_types(
    fake_data
):
    """
    G: Given that users want to generate fake data based on a DataFrame schema
    W: When the function generate_fake_data_from_schema() with a predefined
       schema
    T: Then the elements returned in the list must have a set of expected
       data types based on the input schema provided
    """

    # Extracting data types of the generated fake data
    fake_data_types = [[type(element) for element in row] for row in fake_data]

    # Getting a sample row to compare
    row_data_types = fake_data_types[0]

    # Creating an expected list of data types
    expec_types = [str, int, int, Decimal, float, float, bool, date, datetime]

    assert row_data_types == expec_types


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.generate_fake_dataframe
def test_function_generate_fake_dataframe_returns_a_spark_dataframe_object(
    df_fake
):
    """
    G: Given users want to generate a fake Spark DataFrame object
    W: When the function generate_fake_dataframe() is called
    T: Then the return must be a Spark DataFrame object
    """

    assert type(df_fake) is DataFrame


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.compare_dataframe_schemas
def test_compare_schemas_returns_true_when_comparing_equal_schemas(df_fake):
    """
    G: Given that users want to compare two Spark DataFrame's schemas
    W: When the function compare_dataframe_schemas() is called with two Spark
       DataFrames with the same schema
    T: Then the return must be True
    """

    assert compare_dataframe_schemas(df1=df_fake, df2=df_fake)


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.compare_dataframe_schemas
def test_compare_schemas_returns_false_when_comparing_different_schemas(
    spark_session
):
    """
    G: Given that users want to compare two Spark DataFrame's schemas
    W: When the function compare_dataframe_schemas() is called with two Spark
       DataFrames with different schema
    T: Then the return must be False
    """

    # Defining the attributes for the first fake DataFrame
    fake_schema_1 = [
        {
            "Name": "id",
            "Type": "string",
            "nullable": True
        },
        {
            "Name": "value",
            "Type": "integer",  # This is the difference!
            "nullable": True
        }
    ]

    # Defining the attributes for the second fake DataFrame
    fake_schema_2 = [
        {
            "Name": "id",
            "Type": "string",
            "nullable": True
        },
        {
            "Name": "value",
            "Type": "string",
            "nullable": True
        }
    ]

    # Creating fake DataFrames
    df_fake_1 = generate_fake_dataframe(spark_session, fake_schema_1)
    df_fake_2 = generate_fake_dataframe(spark_session, fake_schema_2)

    # Comparing both DataFrames
    assert not compare_dataframe_schemas(df1=df_fake_1, df2=df_fake_2)


@pytest.mark.tester
@pytest.mark.dataframes
@pytest.mark.compare_dataframe_schemas
def test_compare_schemas_nullable_info_return_false_when_nullable_is_diff(
    spark_session
):
    """
    G: Given that users want to compare two Spark DataFrame's schemas
    W: When the function compare_dataframe_schemas() is called with two Spark
       DataFrames with schemas having the same attributes and the same data
       types but different nullable flag (considering the function is called
       with flag compare_nullable_info=True)
    T: Then the return must be False
    """

    # Defining the attributes for the first fake DataFrame
    fake_schema_1 = [
        {
            "Name": "id",
            "Type": "string",
            "nullable": True
        },
        {
            "Name": "value",
            "Type": "integer",
            "nullable": True
        }
    ]

    # Defining the attributes for the second fake DataFrame
    fake_schema_2 = [
        {
            "Name": "id",
            "Type": "string",
            "nullable": True
        },
        {
            "Name": "value",
            "Type": "integer",
            "nullable": False  # This is the difference!
        }
    ]

    # Creating fake DataFrames
    df_fake_1 = generate_fake_dataframe(spark_session, fake_schema_1)
    df_fake_2 = generate_fake_dataframe(spark_session, fake_schema_2)

    # Comparing both DataFrames
    assert not compare_dataframe_schemas(
        df1=df_fake_1,
        df2=df_fake_2,
        compare_nullable_info=True
    )



"""Test cases for features defined on utils.py module.

This file handles the definition of all test cases for testing everything on
sparksnake's utils module.

___
"""

# Importing libraries
import pytest
import logging

from sparksnake.utils.log import log_config


@pytest.mark.utils
def test_log_config_function_returns_a_logger_object():
    """
    G: Given that users want to obtain a preconfigured logger object
    W: When the log_config functions is called from sparksnake.utils module
    T: Then the returned object must be a logging.Logger object
    """

    logger = log_config()
    assert type(logger) == logging.Logger



"""Package setup script.

This script handles everything about package publishing on PyPI.
"""

# Importing libraries
from setuptools import setup, find_packages

# Reading README.md for project description
with open("README.md", "r", encoding='utf-8') as f:
    __long_description__ = f.read()

# Setting up package information
setup(
    name='sparksnake',
    version='0.1.15',
    author='Thiago Panini',
    author_email='panini.development@gmail.com',
    packages=find_packages(),
    install_requires=[
        "pyspark"
    ],
    license='MIT',
    description="Improving the development of Spark applications deployed as "
                "jobs on AWS services like Glue and EMR",
    long_description=__long_description__,
    long_description_content_type="text/markdown",
    url='https://github.com/ThiagoPanini/sparksnake',
    keywords='Cloud, AWS, Python, Spark, pyspark',
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: Portuguese (Brazilian)",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    python_requires=">=3.0.0"
)



"""Helps user to create and configure their logger object

This module can be used to improve observability in log messages by providing
and easy way to create and retrieve logger objects with a given configuration.

___
"""

# Importing libraries
import logging


# Creating and configuring logs
def log_config(
    logger_name: str = __file__,
    logger_level: int = logging.INFO,
    logger_date_format: str = "%Y-%m-%d %H:%M:%S"
) -> logging.Logger:
    """Logger object configuration.

    This function can be used in any Python application in order to improve log
    proccesses and enhance software observability. It uses the logging built-in
    Python package to retrieve users a basic logger object with a given
    configuration.

    Examples:
        ```python
        # Getting and configuring the logger object
        logger = log_config(logger)
        ```

    Args:
        logger_name (str): A name for the logger object
        logger_level (int): The type of the logger object
        logger_date_format (str): Message format of the logger object

    Returns:
        A preconfigured logger object
    """

    # Creating a logger object and setting its level
    logger = logging.getLogger(logger_name)
    logger.setLevel(logger_level)

    # Setting the format message
    log_format = "%(levelname)s;%(asctime)s;%(filename)s;"
    log_format += "%(lineno)d;%(message)s"
    formatter = logging.Formatter(log_format,
                                  datefmt=logger_date_format)

    # Setting up the stream handler
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    return logger



"""Test cases for features defined on manager.py module.

This file handles the definition of all test cases for testing SparkETLManager
class and its features. The idea is to isolate a test script for testing
sparksnake features delivered for users who want to develop Spark applications
in their local environment through a SparkETLManager class object with
mode="local".

___
"""

# Importing libraries
import pytest

from pyspark.sql.functions import expr
from pyspark.sql.types import StringType, DateType, TimestampType
from pyspark.sql.utils import AnalysisException

from sparksnake.manager import SparkETLManager


@pytest.mark.spark_manager_local
@pytest.mark.constructor
def test_error_on_initializing_class_with_invalid_mode():
    """
    G: Given that an user wants to initialize a SparkETLManager class object
    W: When the SparkETLManager class is initialized with a mode attribute
    different than the acceptable values
    T: Then a ValueError exception muset be raised
    """

    # Initializing class
    with pytest.raises(ValueError):
        _ = SparkETLManager(mode="invalid_mode")


@pytest.mark.spark_manager_local
@pytest.mark.constructor
def test_error_with_glue_mode_without_argvlist_or_datadict_attributes():
    """
    G: Given that an user wants to initialize a SparkETLManager class object
    W: When the SparkETLManager class is initialized with mode="glue" but
    there's no argv_list or data_dict_attributes passed by the user
    T: Then a TypeError exception muset be raised
    """

    # Initializing class
    with pytest.raises(TypeError):
        _ = SparkETLManager(mode="glue")


@pytest.mark.spark_manager_local
@pytest.mark.date_transform
def test_casting_date_column_with_date_transform_method(
    df_fake,
    spark_manager_local,
    date_col="date_string_field",
    date_col_type="date",
    date_format="yyyy-MM-dd"
):
    """
    G: Given that an user has a string column in a DataFrame that has date
    information and they need to cast it to date
    W: When the date_transform method is called with the following parameters:
        - date_col="name-of-the-string-column"
        - date_col_type="date"
        - date_format="yyyy-MM-dd"
        - cast_string_to_date=True
    T: Then the "name-of-the-string-column" column on the resulting DataFrame
    must have the DateType type
    """

    # Casting a source date column to string to test the feature
    df_fake_tmp = df_fake.withColumn(
        "date_string_field",
        expr("cast(date_field AS STRING)")
    )

    # Calling the method for casting a string field to date
    df_fake_prep = spark_manager_local.date_transform(
        df=df_fake_tmp,
        date_col=date_col,
        date_col_type=date_col_type,
        date_format=date_format,
        cast_string_to_date=True
    )

    # Extracting data type BEFORE casting
    dtype_pre_casting = df_fake_tmp.schema[date_col].dataType

    # Extracting data type AFTER casting
    dtype_pos_casting = df_fake_prep.schema[date_col].dataType

    # Asserting the casting
    assert dtype_pre_casting == StringType()
    assert dtype_pos_casting == DateType()


@pytest.mark.spark_manager_local
@pytest.mark.date_transform
def test_casting_timestamp_column_with_date_transform_method(
    df_fake,
    spark_manager_local,
    date_col="timestamp_string_field",
    date_col_type="timestamp",
    date_format="yyyy-MM-dd HH:mm:ss"
):
    """
    G: Given that an user has a string column in a DataFrame that has timestamp
    information and they need to cast it to timestamp
    W: When the date_transform method is called with the following parameters:
        - date_col="name-of-the-string-column"
        - date_col_type="timestamp"
        - date_format="yyyy-MM-dd HH:mm:ss"
        - cast_string_to_date=True
    T: Then the "name-of-the-string-column" column on the resulting DataFrame
    must have the TimestampType type
    """

    # Casting a source date column to string to test the feature
    df_fake_tmp = df_fake.withColumn(
        "timestamp_string_field",
        expr("cast(timestamp_field AS STRING)")
    )

    # Calling the method for casting a string field to date
    df_fake_prep = spark_manager_local.date_transform(
        df=df_fake_tmp,
        date_col=date_col,
        date_col_type=date_col_type,
        date_format=date_format,
        cast_string_to_date=True
    )

    # Extracting data type BEFORE casting
    dtype_pre_casting = df_fake_tmp.schema[date_col].dataType

    # Extracting data type AFTER casting
    dtype_pos_casting = df_fake_prep.schema[date_col].dataType

    # Asserting the casting
    assert dtype_pre_casting == StringType()
    assert dtype_pos_casting == TimestampType()


@pytest.mark.spark_manager_local
@pytest.mark.date_transform
def test_error_on_casting_date_column_with_invalid_date_col_type(
    df_fake,
    spark_manager_local,
    date_col="date_field",
    date_col_type="invalid_type",
    date_format="yyyy-MM-dd"
):
    """
    G: Given that an user has a string column in a DataFrame that has date
    information and they need to cast it to date
    W: When the date_transform method is called with an invalid date_col_type
    attribute (e.g. something different from "date" or "timestamp")
    T: Then an ValueError must be raised
    """

    # Asserting execption raising
    with pytest.raises(ValueError):
        _ = spark_manager_local.date_transform(
            df=df_fake,
            date_col=date_col,
            date_col_type=date_col_type,
            date_format=date_format,
            cast_string_to_date=True
        )


@pytest.mark.spark_manager_local
@pytest.mark.date_transform
def test_error_on_casting_date_column_with_invalid_column_name(
    df_fake,
    spark_manager_local,
    date_col="invalid_column_name",
    date_col_type="date",
    date_format="yyyy-MM-dd HH:mm:ss"
):
    """
    G: Given that an user has a string column in a DataFrame that has date
    information and they need to cast it to date
    W: When the date_transform method is called with a column name that
    doesn't exist on DataFrame
    T: Then an AnalysisException must be raised
    """

    # Asserting execption raising
    with pytest.raises(AnalysisException):
        _ = spark_manager_local.date_transform(
            df=df_fake,
            date_col=date_col,
            date_col_type=date_col_type,
            date_format=date_format,
            cast_string_to_date=True
        )


@pytest.mark.spark_manager_local
@pytest.mark.date_transform
def test_correct_col_name_after_extracting_year_info_from_date(
    spark_manager_local,
    df_fake,
    date_col="date_field"
):
    """
    G: Given that an user wants to add a new column to an existing DataFrame
       with year information based on a date column
    W: When the date_transform() method is executed with kwarg year=True
    T: Then there might be a new DataFrame column named "year_{date_col}"
       where {date_col} is the name of the target date column used to get the
       year
    """

    # Calling the date_transform() method to extract year info from a date
    df_fake_year = spark_manager_local.date_transform(
        df=df_fake,
        date_col=date_col,
        year=True
    )

    assert f"year_{date_col}" in df_fake_year.schema.fieldNames()


@pytest.mark.date_transform
def test_correct_col_value_after_extracting_year_info_from_date(
    spark_manager_local,
    df_fake,
    date_col="date_field"
):
    """
    G: Given that an user wants to add a new column to an existing DataFrame
       with year information based on a date column
    W: When the date_transform() method is executed with kwarg year=True
    T: Then the new year_{date_col} field must have the expected year
       information extracted from {date_col}
    """

    # Calling the date_transform() method to extract year info from a date
    df_fake_year = spark_manager_local.date_transform(
        df=df_fake,
        date_col=date_col,
        year=True
    )

    # Getting the expected field value
    expected_value = df_fake.select(date_col).take(1)[0][0].year

    # Getting the generated value
    current_value = df_fake_year.select(f"year_{date_col}").take(1)[0][0]

    assert current_value == expected_value


@pytest.mark.spark_manager_local
@pytest.mark.date_transform
def test_correct_col_names_after_extracting_all_date_information(
    spark_manager_local,
    df_fake,
    date_col="date_field"
):
    """
    G: Given that an user wants to add a new column to an existing DataFrame
       with all possible date information based on a date column
    W: When the date_transform() method is executed with kwargs year=True,
       quarter=True, month=True, dayofmonth=True, dayofweek=True,
       dayofyear=True, weekofyear=True
    T: Then, for each aforementioned kwarg, there might be a new column
       named "{kwarg}_{date_col}" where {kwarg} is the date information
       to be extracted and {date_col} is the name of the target date column
    """

    # Calling the date_transform() method to extract year info from a date
    df_fake_date = spark_manager_local.date_transform(
        df=df_fake,
        date_col=date_col,
        year=True,
        quarter=True,
        month=True,
        dayofmonth=True,
        dayofweek=True,
        dayofyear=True,
        weekofyear=True
    )

    # Extracting field names from the new DataFrame
    new_field_names = df_fake_date.schema.fieldNames()

    # Creating a list of expected field names to be included
    date_kwargs = ["year", "quarter", "month", "dayofmonth", "dayofweek",
                   "dayofyear", "weekofyear"]
    expected_field_names = [f"{d}_{date_col}" for d in date_kwargs]

    assert all(d in new_field_names for d in expected_field_names)


@pytest.mark.spark_manager_local
@pytest.mark.agg_data
def test_correct_col_names_after_aggregating_data_with_all_possible_functions(
    spark_manager_local,
    df_fake,
    spark_session,
    numeric_col="integer_field",
    group_by="boolean_field"
):
    """
    G: Given that an user wants to aggregate data from a DataFrame
    W: When the agg_data() method is called to aggregate data using all
       possible functions available on method
    T: Then there might new columns with named matching the pattern
       {function}_{numeric_col} where {function} is the aggreaget function flag
       passed on kwargs and {numeric_col} is the numeric col parameter passed
       on method call
    """

    # Creating a kwargs dict with all possible functions to aggregate
    agg_functions = ["sum", "mean", "max", "min", "count", "variance",
                     "stddev", "kurtosis", "skewness"]
    kwargs_dict = {f: True for f in agg_functions}

    # Aggregating data
    df_agg = spark_manager_local.agg_data(
        spark_session=spark_session,
        df=df_fake,
        numeric_col=numeric_col,
        group_by=group_by,
        round_result=True,
        n_round=2,
        **kwargs_dict
    )

    # Creating a list of expected column names to be in df_agg DataFrame
    expected_columns = [f"{f}_{numeric_col}" for f in agg_functions]

    assert all(c in df_agg.schema.fieldNames() for c in expected_columns)


@pytest.mark.spark_manager_local
@pytest.mark.agg_data
def test_columns_on_groupby_list_are_part_of_schema_after_aggregating_data(
    spark_manager_local,
    df_fake,
    spark_session,
    numeric_col="integer_field",
    group_by=["string_field", "boolean_field"]
):
    """
    G: Given that an user wants to aggregate data from a DataFrame
    W: When the agg_data() method is called to aggregate data using a list of
       columns on group_by argument and any agg function
    T: Then the group_by list must be on the returned DataFrame schema
    """

    # Aggregating data
    df_agg = spark_manager_local.agg_data(
        spark_session=spark_session,
        df=df_fake,
        numeric_col=numeric_col,
        group_by=group_by,
        sum=True
    )

    # Checking if group_by list are in schema
    assert all(c in df_agg.schema.fieldNames() for c in group_by)


@pytest.mark.spark_manager_local
@pytest.mark.agg_data
def test_aggregating_data_with_sum_function_returns_expected_value(
    spark_manager_local,
    df_fake,
    spark_session,
    numeric_col="integer_field",
    group_by="boolean_field"
):
    """
    G: Given that an user wants to aggregate data from a DataFrame
    W: When the agg_data() method is called to aggregate data with sum
    T: Then sum_{numeric_col} column value must match the expected
    """

    # Aggregating data
    df_agg = spark_manager_local.agg_data(
        spark_session=spark_session,
        df=df_fake,
        numeric_col=numeric_col,
        group_by=group_by,
        sum=True
    )

    # Collecting rows for the raw DataFrame to be tested
    value_rows = df_fake.where(expr("boolean_field == True"))\
        .select(numeric_col).collect()

    # Applying the sum on the result rows using Python
    expected_result = sum([r[0] for r in value_rows])

    # Collecting the result from the DataFrame after agg_data() method
    agg_result = df_agg.where(expr("boolean_field == True"))\
        .select(f"sum_{numeric_col}").collect()[0][0]

    assert agg_result == expected_result


@pytest.mark.spark_manager_local
@pytest.mark.agg_data
def test_error_on_calling_agg_data_method_with_invalid_numeric_col_reference(
    spark_manager_local,
    df_fake,
    spark_session,
    numeric_col="invalid_col",
    group_by="boolean_field"
):
    """
    G: Given that an user wants to aggregate data from a DataFrame
    W: When the agg_data() method is called with a column reference for
       numeric_col argument that doesn't exist on the target DataFrame
    T: Then a AnalysisException error must be raised
    """

    # Trying to aggregate with invalid numeric_col argument
    with pytest.raises(AnalysisException):
        _ = spark_manager_local.agg_data(
            spark_session=spark_session,
            df=df_fake,
            numeric_col=numeric_col,
            group_by=group_by
        )


@pytest.mark.spark_manager_local
@pytest.mark.agg_data
def test_error_on_calling_agg_data_method_with_invalid_group_by_reference(
    spark_manager_local,
    df_fake,
    spark_session,
    numeric_col="integer_field",
    group_by="invalid_col"
):
    """
    G: Given that an user wants to aggregate data from a DataFrame
    W: When the agg_data() method is called with a column reference for
       group_by argument that doesn't exist on the target DataFrame
    T: Then a AnalysisException error must be raised
    """

    # Trying to aggregate with invalid numeric_col argument
    with pytest.raises(AnalysisException):
        _ = spark_manager_local.agg_data(
            spark_session=spark_session,
            df=df_fake,
            numeric_col=numeric_col,
            group_by=group_by
        )


@pytest.mark.spark_manager_local
@pytest.mark.add_partition_column
def test_correct_col_name_after_adding_partition_with_add_partition_method(
    spark_manager_local,
    df_fake,
    partition_name="execution_date",
    partition_value=0
):
    """
    G: Given that an user wants to add a new partition column to a DataFrame
    W: When the add_partition_column() method is called
    T: Then the partition column name must exists on the returned DataFrame
    """

    # Adding a partition column
    df_fake_partitioned = spark_manager_local.add_partition_column(
        df=df_fake,
        partition_name=partition_name,
        partition_value=partition_value
    )

    assert partition_name in df_fake_partitioned.schema.fieldNames()


@pytest.mark.spark_manager_local
@pytest.mark.add_partition_column
def test_correct_col_value_after_adding_partition_with_add_partition_method(
    spark_manager_local,
    df_fake,
    partition_name="execution_date",
    partition_value=0
):
    """
    G: Given that an user wants to add a new partition column to a DataFrame
    W: When the add_partition_column() method is called
    T: Then the partition column value must match the partition_value argument
    """

    # Adding a partition column
    df_fake_partitioned = spark_manager_local.add_partition_column(
        df=df_fake,
        partition_name=partition_name,
        partition_value=partition_value
    )

    assert df_fake_partitioned.select(partition_name).take(1)[0][0] == 0


@pytest.mark.spark_manager_local
@pytest.mark.add_partition_column
def test_error_on_calling_add_partition_method_with_invalid_partition_name(
    spark_manager_local,
    df_fake,
    partition_name=0,
    partition_value=0
):
    """
    G: Given that an user wants to add a new partition column to a DataFrame
    W: When the add_partition_column() method is called with an invalid value
       for partition_name argument (e.g. an integer number instead of a string
       column name reference)
    T: Then an Exception must be raised
    """

    # Trying to execute the method and caughting exception
    with pytest.raises(Exception):
        _ = spark_manager_local.add_partition_column(
            df=df_fake,
            partition_name=partition_name,
            partition_value=partition_value
        )


@pytest.mark.spark_manager_local
@pytest.mark.repartition_dataframe
def test_decreasing_dataframe_partitions_with_repartition_method(
    spark_manager_local,
    df_fake
):
    """
    G: Given that an user wants to decrease the number of partitions in a
       Spark DataFrame
    W: When the repartition_dataframe() method is called with a number of
       desired partitions on num_partition arg that is LESS THAN the current
       number of DataFrame partitions
    T: Then the returned DataFrame must have the desired number of partitions
    """

    # Getting the current partitions number and setting a desired number
    current_partitions = df_fake.rdd.getNumPartitions()
    partitions_to_set = current_partitions // 2

    # Calling the repartition method
    df_repartitioned = spark_manager_local.repartition_dataframe(
        df=df_fake,
        num_partitions=partitions_to_set
    )

    assert df_repartitioned.rdd.getNumPartitions() == partitions_to_set


@pytest.mark.spark_manager_local
@pytest.mark.repartition_dataframe
def test_increasing_dataframe_partitions_with_repartition_method(
    spark_manager_local,
    df_fake
):
    """
    G: Given that an user wants to increase the number of partitions in a
       Spark DataFrame
    W: When the repartition_dataframe() method is called with a number of
       desired partitions on num_partition arg that is MORE THAN the current
       number of DataFrame partitions
    T: Then the returned DataFrame must have the desired number of partitions
    """

    # Getting the current partitions number and setting a desired number
    current_partitions = df_fake.rdd.getNumPartitions()
    partitions_to_set = current_partitions * 2

    # Calling the repartition method
    df_repartitioned = spark_manager_local.repartition_dataframe(
        df=df_fake,
        num_partitions=partitions_to_set
    )

    assert df_repartitioned.rdd.getNumPartitions() == partitions_to_set


@pytest.mark.spark_manager_local
@pytest.mark.repartition_dataframe
def test_trying_to_repartition_dataframe_with_current_number_of_partitions(
    spark_manager_local,
    df_fake
):
    """
    G: Given that an user wants to change the number of partitions in a
       Spark DataFrame
    W: When the repartition_dataframe() method is called with a number of
       desired partitions on num_partition arg that is EQUAL to the current
       number of DataFrame partitions
    T: Then no repartition operation should be executed and the returned
       DataFrame must have the same number of partitions as before
    """

    # Getting the current partitions number and setting a desired number
    current_partitions = df_fake.rdd.getNumPartitions()
    partitions_to_set = current_partitions

    # Calling the repartition method
    df_repartitioned = spark_manager_local.repartition_dataframe(
        df=df_fake,
        num_partitions=partitions_to_set
    )

    assert df_repartitioned.rdd.getNumPartitions() == partitions_to_set


@pytest.mark.spark_manager_local
@pytest.mark.repartition_dataframe
def test_trying_to_repartition_dataframe_with_negative_number_of_partitions(
    spark_manager_local,
    df_fake
):
    """
    G: Given that an user wants to change the number of partitions in a
       Spark DataFrame
    W: When the repartition_dataframe() method is called with a negative
       number of desired partitions on num_partition argument
    T: Then no operation should be done and the same DataFrame must be
       returned with no changes in its number of partitions
    """

    # Getting the current partitions number and setting a desired number
    current_partitions = df_fake.rdd.getNumPartitions()
    partitions_to_set = -1

    # Calling the repartition method
    df_repartitioned = spark_manager_local.repartition_dataframe(
        df=df_fake,
        num_partitions=partitions_to_set
    )

    assert df_repartitioned.rdd.getNumPartitions() == current_partitions



"""Managing Spark operations in AWS services that uses it in ETL jobs.

This module provides Spark features do be applied in Spark applications to
be deployed locally or using AWS services like Glue and EMR.

___
"""

# Importing libraries
from time import sleep

from sparksnake.utils.log import log_config

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import expr, lit
from pyspark.sql.utils import AnalysisException

try:
    from sparksnake.glue import GlueJobManager as ManagerClass
except ImportError:
    class ManagerClass:
        pass

# Setting up a logger object
logger = log_config(logger_name=__file__)


class SparkETLManager(ManagerClass):
    """Puts together all Spark features used in ETL jobs in AWS.

    This class provides an easy and fast way for users to improve and
    enhance the development of their Spark applications. The class makes it
    possible through a series of features (attributes and methods) specially
    coded for making the development journey something really simple and fun.

    According to the AWS service to be used by the user to run their Spark
    application, this class can be configured through its "mode" attribute.
    Technically, it means that the "mode" attribute literally handles all
    class inheritance processes based on target AWS service to be used.

    So, users can have available special attributes and methods for improving
    their development process wherever they are creating a Spark application
    for running on AWS Glue, Amazon EMR or even locally.

    Example: A basic usage example of class `SparkETLManager` with mode="glue"
        ```python
        # Importing class
        from sparksnake.manager import SparkETLManager

        # Defining job arguments
        ARGV_LIST = ["JOB_NAME", "S3_OUTPUT_PATH"]

        # Defining dictionary of data sources to be used on job
        DATA_DICT = {
            "orders": {
                "database": "ra8",
                "table_name": "orders",
                "transformation_ctx": "dyf_orders"
            },
            "customers": {
                "database": "ra8",
                "table_name": "customers",
                "transformation_ctx": "dyf_customers",
                "push_down_predicate": "anomesdia=20221201",
                "create_temp_view": True,
                "additional_options": {
                    "compressionType": "lzo"
                }
            }
        }

        # Creating a class object on initializing a glue job
        spark_manager = SparkETLManager(
            mode="glue",
            argv_list=ARGV_LIST,
            data_dict=DATA_DICT
        )

        spark_manager.init_job()

        # Getting all DataFrames Spark based on data_dict provided
        dfs_dict = spark_manager.generate_dataframes_dict()

        # Indexing a DataFrame from the dictionary
        df_orders = dfs_dict["orders"]

        # Dropping a partition on S3 (if exists)
        spark_manager.drop_partition(
            s3_partition_uri="s3://some-bucket-name/some-table-name/partition/"
        )

        # Adding a partition column into the DataFrame
        df_orders_partitioned = spark_manager.add_partition_column(
            partition_name="anomesdia",
            partition_value="20230101"
        )

        # Applying a repartition method for storage optimization
        df_orders_repartitioned = spark_manager.repartition_dataframe(
            df=df_orders_partitioned,
            num_partitions=10
        )

        # Writing data on S3 and cataloging it on Data Catalog
        spark_manager.write_data_to_catalog(df=df_orders_repartitioned)

        # Job commit
        spark_manager.job.commit()
        ```

    Args:
        mode (string):
            Operation mode for the class. It handles inheritance from other
            classes based on this library so the `SparkETLManager` class can
            expand its features for a Spark application development in
            specific scenarios.
            Acceptable values are: "glue", "emr", "local".

    Tip: The "mode" attribute may not be the only one
        By its own construction, the class `SparkETLManager`inherits attributes
        and methods from other classes in the library. It means that these
        other classes may have its own attributes to be passed on class
        construction.

        For example, when users choose to set the Glue operation mode
        (`mode="glue"`) on starting up the class `SparkETLManager`, a class
        inheritance from `GlueJobManager` is set under the hood. Because of
        that, attributes like `argv_list` and `data_dict` must be passed
        together with `mode` as they are mandatory on `GlueJobManager` class
        construction.

        For a special tip, check the construction of the class to be inherited
        based on mode operation to see which attributes are necessary to pass
        at `SparkETLManager` class start up.
    """

    def __init__(self, mode: str, **kwargs) -> None:
        self.mode = mode.strip().lower()

        # Checking if the operation mode value was passed as expected
        if self.mode not in ("local", "glue", "emr"):
            raise ValueError(f"The attribute mode was set as {self.mode} "
                             "but acceptable values are 'local', 'glue' and "
                             'emr')

        # Looking out for glue operation mode
        if self.mode == "glue":
            # Checking if required args for GlueJobManager were passed
            if "argv_list" not in kwargs or "data_dict" not in kwargs:
                raise TypeError("The operation mode was set as 'glue' but "
                                "'argv_list' and/or 'data_dict' required "
                                "arguments weren't set properly.")

            # Collecting required args for mode="glue"
            argv_list = kwargs["argv_list"]
            data_dict = kwargs["data_dict"]

            # Applying class inheritance for this mode
            try:
                ManagerClass.__init__(self, argv_list=argv_list,
                                      data_dict=data_dict)
            except TypeError as te:
                logger.error("Error on inherting class GlueJobManager. Check "
                             "if your environment has the awsglue libraries "
                             "and try again. If you don't have awsglue libs "
                             "available, you probably want to run sparksnake "
                             "in a local operation mode. If this is the case, "
                             "change the mode attribute to 'local'")
                raise te

            # Logging initialization message
            logger.info("The class was succesfully initialized with Glue "
                        "operation mode and all Glue sparksnake features were "
                        "inherited and are now available")

        # Looking out for local mode
        elif self.mode == "local":
            # Checking if user wants to pass its own SparkSession object
            if "spark" in kwargs and type(kwargs["spark"]) is SparkSession:
                self.spark = kwargs["spark"]
            else:
                logger.info("Creating a SparkSession object to be used in a "
                            "local environment")
                self.spark = SparkSession.builder\
                    .appName("sparksnake-app")\
                    .getOrCreate()

            # Logging initialization message
            logger.info("The class was succesfully initialized with 'local' "
                        "operation mode")

    @staticmethod
    def date_transform(df: DataFrame,
                       date_col: str,
                       date_col_type: str = "date",
                       date_format: str = "yyyy-MM-dd",
                       cast_string_to_date: bool = True,
                       **kwargs) -> DataFrame:
        """Extracting date attributes from a Spark DataFrame date column.

        This method makes it possible to extract multiple date attributes from
        a column in a Spark DataFrame that represents a date or timestamp
        information. To do that, the given date column in the DataFrame should
        be of types DATE, TIMESTAMP. There is a special case where the target
        date column is of STRING data type. In that situtation, the column
        must be parseable as a date so the date attributes can be extracted
        from it.

        The method logic is made for a initial field casting validation
        followed by the extraction of all date attributes according to the user
        input on the method call. The big idea behing this features is the
        ability to provide users a huge DataFrame enrichment with a single
        method call to improve their analytics processes.

        Examples:
            ```python
            # Extracting date attributes from a date column in a Spark df
            df_date_prep = spark_manager.date_transform(
                df=df_raw,
                date_col="order_date",
                date_col_type="timestamp",
                year=True,
                month=True,
                dayofmonth=True
            )

            # In the above example, the method will return a new DataFrame
            # with additional columns based on the order_date_content, such as:
            # year_order_date, month_order_date and dayofmonth_order_date
            ```

        Args:
            df (pyspark.sql.DataFrame):
                A target Spark DataFrame for applying the transformation.

            date_col (str):
                A date column name (or parseable string as date) to be used in
                the date extraction process.

            date_col_type (str):
                Reference for data type of `date_col` argument. Acceptable
                values are "date" or "timestamp".

            date_format (str):
                Date format applied in a optional string to date casting.
                It's applicable only if `cast_string_to_date=True`

            cast_string_to_date (bool):
                Enables an automatic casting of the `date_col` column reference
                into a given `date_format`.

        Keyword Args:
            year (bool): Extracts the year of target date column
            quarter (bool): Extracts the quarter of target date column
            month (bool): Extracts the month of target date column
            dayofmonth (bool): Extracts the dayofmonth of target date column
            dayofweek (bool): Extracts the dayofweek of target date column
            weekofyear (bool): Extracts the weekofyear of target date column

        Raises:
            ValueError: Exception raised if the `date_col_type` argument is\
            passed in non acceptable value (e.g. something different of "date"\
            or "timestamp").

        Returns:
            Spark DataFrame with new date columns extracted.
        """

        try:
            # Creating casting expressions based on data type of date_col arg
            date_col_type = date_col_type.strip().lower()
            if cast_string_to_date:
                if date_col_type == "date":
                    casting_expr = f"to_date({date_col},\
                        '{date_format}') AS {date_col}_{date_col_type}"
                elif date_col_type == "timestamp":
                    casting_expr = f"to_timestamp({date_col},\
                        '{date_format}') AS {date_col}_{date_col_type}"
                else:
                    raise ValueError("The data type of date_col_type parameter"
                                     " is invalid. Acceptable values are "
                                     "'date' or 'timestamp'")

                # Applying a select expression for casting data if applicable
                df = df.selectExpr(
                    "*",
                    casting_expr
                ).drop(date_col)\
                    .withColumnRenamed(f"{date_col}_{date_col_type}", date_col)

        except ValueError as ve:
            logger.error(ve)
            raise ve

        except AnalysisException as ae:
            logger.error("Analysis error on trying to cast the column "
                         f"{date_col} using the expression {casting_expr}. "
                         "Maybe this column doesn't exist on the DataFrame. "
                         f"Check the error treceback for more details: {ae}")
            raise ae

        # Creating a list of all possible date attributes to be extracted
        possible_date_attribs = ["year", "quarter", "month", "dayofmonth",
                                 "dayofweek", "dayofyear", "weekofyear"]

        # Iterating over all possible attributes and extracting date attribs
        for attrib in possible_date_attribs:
            # Add a new column only if attrib is in kwargs
            if attrib in kwargs and bool(kwargs[attrib]):
                df = df.withColumn(
                    f"{attrib}_{date_col}", expr(f"{attrib}({date_col})")
                )

        return df

    @staticmethod
    def agg_data(spark_session: SparkSession,
                 df: DataFrame,
                 numeric_col: str,
                 group_by: str or list,
                 round_result: bool = False,
                 n_round: int = 2,
                 **kwargs) -> DataFrame:
        """Extracts statistical attributes based on a group by operation.

        This method makes it possible to extract multiple statistical
        aggregations based in a numerical column and a set of columns to be
        grouped by. With this feature, users can configure complex aggregations
        in a single method call in order to enhance their analysis.

        Examples:
            ```python
            # Creating a new special and aggregated DataFrame
            df_stats = spark_manager.agg_data(
                spark_session=spark,
                df=df_orders,
                numeric_col="order_value",
                group_by=["order_id", "order_year"],
                sum=True,
                mean=True,
                max=True,
                min=True
            )

            # In the above example, the method will return a new DataFrame with
            # the following columns:
            # order_id e order_year (group by)
            # sum_order_value (sum of order_value column)
            # mean_order_value (average of order_value column)
            # max_order_value (max value of order_value column)
            # min_order_value (min value of order_value column)
            ```

        Args:
            spark_session (pyspark.sql.SparkSession):
                A SparkSession object to be used to run SparkSQL query for
                grouping data

            df (pyspark.sql.DataFrame):
                A target Spark DataFrame for applying the transformation

            numeric_col (str):
                A numeric column name on the target DataFrame to be used as
                target of aggregation process

            group_by (str or list):
                A column name or a list of columns used as group categories
                on the aggregation process

            round_result (bool):
                Enables rounding aggregation results on each new column

            n_round (int):
                Defines the round number on rounding. Applied only if
                `round_result=True`

        Tip: About keyword arguments
            In order to provide a new feature that is capable to put together
            the extraction of multiple statistical attributes with a single
            line of code, a special list of pyspark functions were selected
            as acceptable functions to be called on the aggregation process.

            It means that if users wants to apply an aggregation on the
            target DataFrame and extract the sum, the mean, the minimum and
            the maximum value of a given numeric column, they must pass
            keyword arguments as following: `sum=True`, `mean=True`,
            `min=True` and `max=True`.

            All acceptable keyword arguments (pyspark functions) can be
            found right below:

        Keyword Args:
            sum (bool): Extracts the sum of a given numeric column
            mean (bool): Extracts the mean of a given numeric column
            max (bool): Extracts the max of a given numeric column
            min (bool): Extracts the min of a given numeric column
            countDistinct (bool): Extracts the count distinct value\
                of a given numeric column
            variance (bool): Extracts the variance of a given numeric column
            stddev (bool): Extracts the standard deviation of a given numeric\
                column
            kurtosis (bool): Extracts the kurtosis of a given numeric column
            skewness (bool): Extracts the skewness of a given numeric column

        Returns:
            A new Spark DataFrame with new statistical columns based on the\
            aggregation configured by user on method call.

        Raises:
            Exception: Generic exception raised when failed to execute the\
            SparkSQL query for extracting the stats from the DataFrame.
        """

        # Joining all group by columns in a single string to make agg easier
        if type(group_by) == list and len(group_by) > 1:
            group_by = ",".join(group_by)

        # Creating a Spark temporary table for grouping data using SparkSQL
        df.createOrReplaceTempView("tmp_extract_aggregate_statistics")

        possible_functions = ["sum", "mean", "max", "min", "count",
                              "variance", "stddev", "kurtosis", "skewness"]
        try:
            # Iterating over the attributes to build a single aggregation expr
            agg_query = ""
            for f in possible_functions:
                if f in kwargs and bool(kwargs[f]):
                    if round_result:
                        agg_function = f"round({f}({numeric_col}), {n_round})"
                    else:
                        agg_function = f"{f}({numeric_col})"

                    agg_query += f"{agg_function} AS {f}_{numeric_col},"

            # Dropping the last comma on the expression
            agg_query = agg_query[:-1]

            # Building the final query to be executed
            final_query = f"""
                SELECT
                    {group_by},
                    {agg_query}
                FROM tmp_extract_aggregate_statistics
                GROUP BY
                    {group_by}
            """

            return spark_session.sql(final_query)

        except AnalysisException as ae:
            logger.error("Error on trying to aggregate data from DataFrame "
                         f"using the following query:\n {final_query}. "
                         "Possible reasons are: missing to pass group_by "
                         "parameter or numeric_col argument doesn't exists "
                         f"on the DataFrame. Exception: {ae}")
            raise ae

    @staticmethod
    def add_partition_column(df: DataFrame,
                             partition_name: str,
                             partition_value) -> DataFrame:
        """Adding a "partition" column on a Spark DataFrame.

        This method is responsible for adding a new column on a target Spark
        DataFrame to be considered as a table partition. In essence, this
        method uses the native pyspark `.withColumn()` method for adding a
        new column to the DataFrame using a name for the partition column
        (partition_name) and its value (partition_value).

        Examples
            ```python
            # Defining partition information
            partition_name = "anomesdia"
            partition_value = int(datetime.strftime('%Y%m%d'))

            # Adding a partition column to the DataFrame
            df_partitioned = spark_manager.add_partition_column(
                df=df_orders,
                partition_name=partition_name,
                partition_value=partition_value
            )

            # The method returns a new DataFrame with a new column
            # referenced by "anomesdia" and its value referenced by
            # the datetime library
            ```

        Args:
            df (pyspark.sql.DataFrame): A target Spark DataFrame.
            partition_name (str): Column name to be added on the DataFrame.
            partition_value (Any): Value for the new column to be added.

        Returns:
            A Spark DataFrame with the new column added.

        Raises:
            Exception: A generic exception raised on failed to execute the\
            method `.withColumn()` for adding the new column.
        """

        logger.info("Adding a new partition column to the DataFrame "
                    f"({partition_name}={str(partition_value)})")
        try:
            df_partitioned = df.withColumn(partition_name,
                                           lit(partition_value))
            return df_partitioned

        except Exception as e:
            logger.error("Error on adding a partition colum to the DataFrame "
                         f"using the .withColumn() method. Exception: {e}")
            raise e

    @staticmethod
    def repartition_dataframe(df: DataFrame, num_partitions: int) -> DataFrame:
        """Repartitioning a Spark DataFrame for optimizing storage.

        This method applies the repartition process in a Spark DataFrame in
        order to optimize its storage on S3. The method has some important
        checks based on each pyspark method to use for repartitioning the
        DataFrame. Take a look at the below tip to learn more.

        Tip: Additional details on method behavior
            The method `repartition_dataframe()` works as follows:

            1. The current number of partitions in the target DataFrame
            is checked
            2. The desired number of partitions passed as a parameter
            is checked

            * If the desired number is LESS than the current number, then the
            method `coalesce()` is executed
            * If the desired number is GREATER than the current one, then the
            method `repartition()` is executed

        Examples:
            ```python
            # Repartitioning a Spark DataFrame
            df_repartitioned = spark_manager.repartition_dataframe(
                df=df_orders,
                num_partitions=10
            )
            ```

        Args:
            df (pyspark.sql.DataFrame): A target Spark Dataframe.
            num_partitions (int): Desider number of partitions.

        Returns:
            A new Spark DataFrame gotten after the repartition process.

        Raises:
            Exception: A generic exception is raised on a failed attempt to\
            run the repartition method (`coalesce()` or `repartition()`) in\
            the given Spark DataFrame.
        """

        # Casting arg to integer to avoid exceptions
        num_partitions = int(num_partitions)

        # Getting the current number of partitions of the given DataFrame
        logger.info("Getting the current number of partition of the DataFrame")
        try:
            current_partitions = df.rdd.getNumPartitions()

        except Exception as e:
            logger.error("Error on collecting the current number of "
                         "using df.rdd.getNumPartitions method. "
                         f"Exception: {e}")
            raise e

        # If the desired partition number if equal to the current number, skip
        if num_partitions == current_partitions:
            logger.warning(f"The current number of partitions "
                           f"({current_partitions}) is equal to the desired "
                           f"number ({num_partitions}). There is no need to "
                           "run any Spark repartition method")
            sleep(0.01)
            return df

        # If the desired number if LESS THAN the current, use coalesce()
        elif num_partitions < current_partitions:
            logger.info("Initializing the repartition process using "
                        f"coalesce(), changing the number of DataFrame "
                        f"partitions from {current_partitions} to "
                        f"{num_partitions}")
            try:
                df_repartitioned = df.coalesce(num_partitions)

            except Exception as e:
                logger.warning("Error on repartitioning using coalesce(). "
                               "The repartition process won't be executed and "
                               "the target DataFrame will be returned without "
                               f"any changes. Exception: {e}")
                return df

        # If the desired number is GREATER THAN the current, use repartition()
        elif num_partitions > current_partitions:
            logger.warning(f"The desired partition number ({num_partitions}) "
                           f"is greater than the current one "
                           f"({current_partitions}) and, because of that, "
                           "the repartitioning operation will be executed "
                           "using the repartition() method which can be "
                           "expensive under the application perspective. As a "
                           "suggestion, check if this is really the expected "
                           "behavior for this operation.")
            try:
                df_repartitioned = df.repartition(num_partitions)

            except Exception as e:
                logger.warning("Error on repartitioning using repartition(). "
                               "The repartition process won't be executed and "
                               "the target DataFrame will be returned without "
                               f"any changes. Exception: {e}")
                return df

        return df_repartitioned
